/*************************************************************************/
/*************************************************************************/
/*                                                                       */
/*  File Name : EdgeList.cc                                              */
/*                                                                       */
/*       Date : May   15, 1999                                           */
/* Updated at : July  10, 1999                                           */
/*                                                                       */
/*     Author : H. Timucin OZDEMIR                                       */
/*                                                                       */
/*************************************************************************/
/**
     EdgeList represents a vehicle duty which can
       1) starts from and ends at the same depot, and
       2) total demand is no more than the maximum capacity of a vehicle.
    The cost is calculated as follows:
      *) Total length of the travel. 
**/
/*************************************************************************/

#include <iostream.h>
#include <iomanip.h>
#include <stdio.h>
#include <stdlib.h>
#include <fstream.h>

#include <gVector.h> 

#include "Edge.h" 
#include "EdgeList.h" 
#include "debugs.h" 
#include "VRPcommons.h" 
#include "TimeWindow.h" 

/*************************************************************************/
/*************************************************************************/
EdgeList::EdgeList() : gVector<Edge>() 
{  myReset();  
   dCostCalculated=0;
   wtime=dCost=tCapacity=0;
}

/*************************************************************************/
/*************************************************************************/
EdgeList::EdgeList(int size) : gVector<Edge>(size) 
{ myReset();wtime=dCost=tCapacity=0; } 

/*************************************************************************/
/*
   methods : gVector<Edge>::addElement()
*/
/*************************************************************************/
EdgeList::EdgeList(Edge& edge) : gVector<Edge>() {  
   wtime=dCost=tCapacity=0;
   addElement(edge); 
   dCostCalculated=0;
}

/*************************************************************************/
/*
   methods : copyFrom()
*/
/*************************************************************************/
EdgeList::EdgeList(const EdgeList& RHS) {  
   wtime=tCapacity=0;
   copyFrom(RHS); 
   dCostCalculated=0;
} 

/*************************************************************************/
/*
   methods : gVector<Edge>::copyFrom()
*/
/*************************************************************************/
void EdgeList::copyFrom(const EdgeList& RHS) {  
   gVector<Edge>::copyFrom(RHS);
   tCapacity=RHS.tCapacity;
   dCost=RHS.dCost;
   wtime=RHS.wtime;
   dCostCalculated=0;
} 

/*************************************************************************/
/*************************************************************************/
void EdgeList::myReset() {
   ecost = 0.0;
   dCostCalculated=0;
}// end of myReset()

/*************************************************************************/
/*
   methods : gVector<Edge>::addElement()
*/
/*************************************************************************/
void EdgeList::addEdge(Edge& edge) { 

   if(!edge.isValid()) { 
     cout << " Edge " << edge << " not valid " << endl; 
     printTabular(cout);
     exit(1); }
   addElement(edge);

   if(getSize()==1) tCapacity=edge.getStartNodeDemand();
   tCapacity+=edge.getEndNodeDemand();

   //Update cost
   dCost += edge.getCost();
   wtime += edge.getWaitingTime();

}// end of addEdge() 

/*************************************************************************/
/*************************************************************************/
Edge *EdgeList::getFirstEdge() const { return getFirstElement(); }

/*************************************************************************/
/*************************************************************************/
Edge *EdgeList::getLastEdge() const  { return getLastElement(); } 

/*************************************************************************/
/*************************************************************************/
Edge *EdgeList::getNthEdge(unsigned indx) const { return getNthElement(indx); } 
/*************************************************************************/
/*************************************************************************/
int EdgeList::getNumberOfEdges() const { return getIndex(); } 

/*************************************************************************/
/*************************************************************************/
EdgeList::getStartNode() const 
{ return (getFirstEdge()!=NULL?getFirstEdge()->getStartNode():-1); }

/*************************************************************************/
/*************************************************************************/
EdgeList::getEndNode() const 
{ return (getLastEdge()!=NULL?getLastEdge()->getEndNode():-1); }

/*************************************************************************/
/*************************************************************************/
vrpcapacity EdgeList::getDemandFromTo(int from, int to,VRProblemPtr prob) const 
{ vrpcapacity ret=0;
  int fr1, i=from;
  if(from>-1&&to>-1) {
   Customer cus;
   EdgePtr edge;
   for(i=from;i<to&&i<getNumberOfEdges();i++) 
     if( (edge=getNthEdge(i))!=NULL) { 
       fr1=edge->getStartNode(); 
       if(prob->getCustomer(fr1,cus)) ret+=cus.getDemand();
     }// end of if
  }// end of if
  return ret;
}// end of getDemandFromTo()

/*************************************************************************/
/** If the start time of the second node is less than the end date of
    the first node of last edge, then this can follow. 
*/
/*************************************************************************/
EdgeList::canFollow(EdgeListPtr RHS, VRProblem* prob) {  

   int ret = (RHS!=NULL? (getStartNode()==RHS->getEndNode()):1);
   if(ret&&RHS!=NULL&&getNumberOfEdges()>0&&RHS->getNumberOfEdges()>0) {
     EdgePtr first = getFirstEdge();
     EdgePtr last  = RHS->getLastEdge();
     EdgeGraphPtr myGraph = prob->getEdgeGraph();
     ret = (first->getStartDateSecondNode()>=
	    (last->getEndDateFirstNode()+myGraph->getTravelTime(last->getStartNode(),first->getEndNode())));
   }// end of if

   return ret;

}// end of canFollow()

/*************************************************************************/
/** If it start from a depot than the start date is the second nodes
    date.
**/
/*************************************************************************/
vrptime EdgeList::getStartDate() const
{
   vrptime ret = VRPTIME_MAX;
   EdgePtr first = getFirstElement();
   if(first!=NULL) 
      ret = (first->doesStartFromDepot()? first->getStartDateSecondNode():
                                          first->getStartDate());
   return ret;

}// end of getStartDate()

/*************************************************************************/
/**
    if it ends at the depot, then the end date is the second nodes
    departure date.
**/
/*************************************************************************/
vrptime EdgeList::getEndDate() const
{
   vrptime ret = VRPTIME_MAX;
   EdgePtr last = getLastElement();
   if(last!=NULL) 
      ret = (last->doesEndAtDepot()? last->getEndDateFirstNode():
                                     last->getEndDate());
   return ret;

}// end of getEndDate()

/*************************************************************************/
/*************************************************************************/
TimeWindow EdgeList::getSpan() const
{ vrptime f = VRPTIME_MAX; 
  vrptime t = VRPTIME_MAX; 
  if(getFirstEdge()!=NULL) f=getFirstEdge()->getStartDate();
  if(getLastEdge()!=NULL)  t=getLastEdge()->getEndDate();
  TimeWindow ret(f,t);
  return ret; 
}// end of getSpan()

/*************************************************************************/
/** Removes the edges which ends at the given node id.  */
/*************************************************************************/
void EdgeList::removeEdgesEndAt(int indx) {
   Edge *edge = NULL;
   int i = getSize() - 1 ;

   // Cost and Edges needs to be reorganized ********
   for(;i>=0;i--)
    if((edge=getNthEdge(i))!=NULL&&edge->getEndNode()==indx) removeElementAt(i);

   myReset();

}// end of removeEdgesEndAt()

/*************************************************************************/
/** removes the edges which starts from the given node id.  */
/*************************************************************************/
void EdgeList::removeEdgesStartFrom(int indx) 
{
   Edge *edge = NULL;
   int i = getSize() - 1;

   // Cost and Edges needs to be reorganized ********
   for(;i>=0;i--)
    if((edge=getNthEdge(i))!=NULL&&edge->getStartNode()==indx) removeElementAt(i);
   myReset();

}// end of removeEdgesStartFrom()

/*************************************************************************/
/*************************************************************************/
void EdgeList::removeFromTo(int from, int to) {
   if( from>=0 && to<getIndex() ) {
      myReset();
      int i=to;
      while(i>=from) { removeElementAt(i); i--; }// end of while
   }// end of if
}// end of removeEdgesEndAt()


/*************************************************************************/
/*
// returns -1 : if the node is not in the list
//         >=0 : index of edge
*/
/*************************************************************************/

EdgeList::doesContainNodeId(int nodeId) const {
   int i=0, ret = -1;
   EdgePtr edge;
   while(i<getSize() && ret==-1) 
      if( (edge = getNthEdge(i))!=NULL && ( edge->getStartNode()==nodeId || 
	  edge->getEndNode()==nodeId ) ) ret = i;
      else i++;
   return ret;
}// end of doessContainNodeId()

/*************************************************************************/
/*
// returns 0 : if the edge is not in the list
//        >0 : if the edge is in the list
*/
/*************************************************************************/

EdgeList::doesContainEdge(int from, int to) const {
   int i=0, ret = 0;
   EdgePtr edge;
   while( i<getSize() && !ret ) 
      if( (edge = getNthEdge(i))!=NULL && edge->getStartNode()==from && 
	  edge->getEndNode()==to ) ret = 1;
      else i++;
   return ret;
}// end of doesContainEdge()

/*************************************************************************/
/*************************************************************************/

EdgeList::doesCommonEdgeExist(EdgeListPtr eList) const
{
   int retVal=1;
   int i=0,j;
   EdgePtr mC;
   EdgePtr C;
   if( eList!=NULL && getNumberOfEdges()>0 )
      for(j=0;retVal&&j<getNumberOfEdges()-1;j++)
      { mC =getNthEdge(j);
	if(mC!=NULL)
  	  for(i=0;retVal&&i<eList->getNumberOfEdges()-1;i++)
	  { C = eList->getNthEdge(i);
            retVal = (C->getEndNode()!=mC->getEndNode());
          }//end of for
      }// end of for
   return retVal;
}// end of doesCommonEdgeExist()

/*************************************************************************/
/*
// returns -1 : if the node is not in the list
//         >=0 : index of edge
*/
/*************************************************************************/

void EdgeList::splitListFromNode( int nodeId, EdgeList& first, 
				  EdgeList& second, VRProblemPtr prob) 
{
   EdgeList _first;
   EdgeList _second;
   int i=0,j,stay=0;
   EdgePtr edge;

   if((i=doesContainNodeId(nodeId))>=0 && i<getIndex() ) {
     j=0;stay=1;
     int currNode = getStartNode();
     vrptime startD = getStartDate();
     for(;j<i&&stay;j++) 
      if((edge=getNthEdge(j))!=NULL) { 
        if( edge->getEndNode()!=nodeId ) { 
	  currNode=edge->getEndNode(); 
	  first.addEdge(*edge); 
	  startD=edge->getStartDateSecondNode();
	} else stay = 0;
      }// end of if
      Edge anEdge(prob->getEdgeGraph(),currNode,getStartNode(),startD);
      first.addEdge(anEdge);
      Customer cId;

     j++;
     if(j<getIndex()) {
       if((edge=getNthEdge(j))!=NULL ) {
         currNode=edge->getStartNode(); 
         startD=getStartDate();
         if(prob->getNodeWithNodeId(currNode,cId)) {
           Edge anEdge( prob->getEdgeGraph(),getStartNode(),currNode,
		        minimum(startD,cId.getStartTime()));
           second.addEdge(anEdge);
           for(;j<getIndex();j++) 
	     if((edge=getNthEdge(j))!=NULL) {
	       startD=anEdge.getStartDateSecondNode();
	       anEdge=*edge;
	       anEdge.setStartDate(startD);
	       second.addEdge(anEdge);
	     }//second.addEdge(*edge);
        }// end of if
      }// end of if
     }// end of if

   } else _first=*this; // end of if

   first = _first;
   second = _second;

}// end of splitListFromNode()

/*************************************************************************/
/**  Equality can be done comparing each element since EdgeList keeps    */
/**  them in time order.                                                 */
/*************************************************************************/
EdgeList::operator==(const EdgeList& RHS) {
   int i, ret = 0;
   ret = (getSize() == RHS.getSize());
   for(i=0;ret && i<getSize();i++) ret=( *getNthEdge(i) == *RHS.getNthEdge(i));
   return ret;
}// end of operator==()

/*************************************************************************/
/**  Assignment operation.  */
/*************************************************************************/
EdgeList& EdgeList::operator=(const EdgeList& RHS) {
   dispose();
   copyFrom(RHS);
   myReset();
   dCostCalculated=0;
   return (*this);
}// end of operator=()

/*************************************************************************/
/*************************************************************************/
void EdgeList::nicePrint(ostream& os) const
{
   int i;
   Edge *curr = NULL; 

   os << " { ";

   for(i=0;i<getNumberOfEdges();i++) 
      if((curr=getNthEdge(i))!=NULL) os << curr->getStartNode() << " -> " ; 
      else os << " Node is NULL " << endl;

   if(curr!=NULL) os << curr->getEndNode();

   os << " } " ; 

   os << " [ ("<<getNumberOfEdges()<<") ";

   for(i=0;i<getNumberOfEdges();i++) 
      if((curr=getNthEdge(i))!=NULL) { curr->nicePrint2(os); os << ";" ; } 
      else os << " NULL " ;

   os << " ] " 
      << " " << getStartDate()
      << ":" << getEndDate()
      << "<" << getDemand() << ">"
      << endl;

}// end of  nicePrint()

/*************************************************************************/
/*************************************************************************/
ostream& operator<<(ostream& os, const EdgeList& RHS) {
   RHS.nicePrint(os);
   return os;
}// end of  operator<<()

/*************************************************************************/
/*************************************************************************/
void EdgeList::printTabular(ostream& os) const {
   int i;
   Edge *curr = NULL; 

   for(i=0;i<getNumberOfEdges();i++) 
     if((curr=getNthEdge(i))!=NULL) curr->printTabular(os);
     else os << " Node is NULL " << endl;

   os << " [ ("<<getNumberOfEdges()<<") "
      << " ] " << getStartDate()
      << ":"   << getEndDate()
      << "<"   << setw(4)  << getDemand() 
      << ","   << setw(10) << getDCost() << ">"
      << endl;

}// end of  printTabular()

/*************************************************************************/
/**
    A sequence is valid if and only if there is no time window
    conflicts.
**/
/*************************************************************************/
EdgeList::isValid(VRProblem *prob, int giveReason) 
{
   int retVal = 0, stay=1,i=1;
   EdgePtr prev=NULL;
   EdgePtr curr=getFirstEdge();
   EdgeGraphPtr myGraph = prob->getEdgeGraph();
   vrptime currTime = (curr!=NULL?curr->getStartDate():0);
   vrpcost tcost = (curr!=NULL?curr->getCost():0);
   vrpcapacity tt = 0;
   Customer cF;
   Customer cT;
   if(prob!=NULL&&curr!=NULL) {
      retVal=1;
      while(curr!=NULL&&stay&&i<getIndex()) {
        if( prob->getCustomer(curr->getStartNode(),cF) && 
            prob->getCustomer(curr->getEndNode(),cT) && 
	    cT.canFollow(cF,currTime)) {
	     currTime = maximum(maximum(cF.getStartTime(),currTime)+cF.getServiceTime()+myGraph->getTravelTime(curr->getStartNode(),curr->getEndNode()),cT.getStartTime());
	 curr=getNthEdge(i); i++;
	 tcost+=curr->getCost();
	 tt+=cT.getDemand();

        } else { stay=0; retVal=0;
		 if(giveReason) 
		    cout << "currTime: " << currTime << endl
		         << "curr: " << *curr << endl 
		         << "cF: " << cF 
		         << "cT: " << cT 
			 << "distance: " 
			 << myGraph->getTravelTime(curr->getStartNode(),curr->getEndNode())
			 <<endl;
	       }

      }// end of while

   }// end of if

   if(retVal&&(tt!=getDemand()||abs(tcost-getDCost())>0.00001||getDemand()>prob->getMaximumCapacity())) { 
     if(giveReason) {
       if(abs(tcost-getDCost())>0.00001)
          cout << "cost value is invalid tcost:" << tcost 
  	       << " eCost: " << getDCost() << " tt: " << tt  
 	       << " getDemand: " << getDemand()<< endl; 
       if(tt!=getDemand())
          cout << "demand value is invalid tt:" << tt 
	       << " getDemand(): " << getDemand() << endl;
       if(getDemand()>prob->getMaximumCapacity())  
          cout << "overflow demand value getDemand(): " << getDemand() 
               << " Maximum V. Capacity: "<< prob->getMaximumCapacity() << endl;
     }// end of if
     retVal=0; 
   }// end of if

   return retVal;
}// end of isValid()

/*************************************************************************/
/** 
    Two edgelists can be merged if:
      total demand is less than the vehicle capacity
*/
/*************************************************************************/
EdgeList::canMerge( EdgeListPtr eList, VRProblem* problem)
{
   int i;
   vrptime startD;  
   vrptime endD;  
   int ret=0;
   Customer cus;

   if(  problem!=NULL && 
       (eList->getDemand()+getDemand())<=problem->getVehicleCapacity() )
   {
     EdgeGraphPtr eGraph = problem->getEdgeGraph();

     if( eList->getEndDate()<getStartDate() )
     {
        EdgePtr last  = eList->getLastEdge();
        EdgePtr first = getFirstEdge();

        // If there is an edge to combine them
        ret = ( last!=NULL && first!=NULL && 
	        eGraph->doesEdgeExist(last->getStartNode(),first->getEndNode(),last->getStartDate()) &&
		(last->getEndDateFirstNode()+eGraph->getTravelTime(last->getStartNode(),first->getEndNode()))<=first->getStartDateSecondNode() );
//	        eGraph->doesEdgeExist(last->getStartNode(), first->getEndNode(), last->getStartDate() ) );

     } else if( eList->getStartDate()>getEndDate() )
            {
               EdgePtr last  = getLastEdge();
               EdgePtr first = eList->getFirstEdge();

               // If there is an edge to combine them
               ret = ( last!=NULL && first!=NULL && 
	        eGraph->doesEdgeExist(last->getStartNode(),first->getEndNode(),last->getStartDate()) &&
		(last->getEndDateFirstNode()+eGraph->getTravelTime(last->getStartNode(),first->getEndNode()))<=first->getStartDateSecondNode() );
	        //       eGraph->doesEdgeExist(last->getStartNode(), first->getEndNode(), last->getStartDate() ) );
            }// end of if

     else {
	    EdgeListPtr rt = EdgeList::merge2EdgeList(this,eList,problem);
	    if(rt!=NULL) { ret=1; delete rt; }
	  }

   }// end of if

   //if(ret) {
   //  ofstream out("canmerge.out",ios::app);
   //  out <<" canMerge " << *this << " with " << *eList << endl;
   //  out.close();
   //}// end of if

   return ret;
      
}// end of canMerge()

/*************************************************************************/
/** Da->Ca->Cb->Da M Da->Cc->Cd->Da ==> Da->Ca->Cb->Cc->Cd->Da OR
                                        Da->Cc->Cd->Ca->Cb->Da
*/
/*************************************************************************/
EdgeList::mergeWith( EdgeListPtr eList ) {
   int i;
   vrptime startD;  
   vrptime endD;  
   int ret=0;

   EdgeGraphPtr eGraph = NULL;
   EdgePtr      mfEdge = getFirstEdge();
   if(mfEdge==NULL) { ret=1; *this=*eList; }
   else eGraph = mfEdge->getEdgeGraph();

   if( eGraph!=NULL && eList->getEndDate()<getStartDate() ) {
        EdgePtr last  = eList->getLastEdge();
        EdgePtr first = getFirstEdge();

        // If there is an edge to combine them
        if( last!=NULL && first!=NULL && 
	    eGraph->doesEdgeExist(last->getStartNode(),
				  first->getEndNode(),last->getStartDate())) 
	{   
	    /*
	    //cout << " 1.mergeWith: " << endl;
	    //cout << " eList " << endl;
	    //eList->printTabular(cout);
	    //cout << " this " << endl;
	    //printTabular(cout);
	    // Becareful, two edges are replaced by one edge.
	    vrpcost dummy0 = first->getCost();
	    vrpcost dummy1 = last->getCost();
            //cout <<" dummy0 " << dummy0 << " dummy1 " << dummy1 << endl;
	    Edge eEdge(eGraph,last->getStartNode(),first->getEndNode(),last->getStartDate());
            //cout <<" dCost " << dCost << " eEdge.getCost() " << eEdge.getCost() << endl;
	    dCost += (eEdge.getCost()-dummy0);
	    //cout << " dCost " << dCost << " tCapacity " << tCapacity ;
	    //cout << " last->getEndNodeDemand() " << last->getEndNodeDemand()
            //     << " first->getStartNodeDemand() " << first->getStartNodeDemand() 
            //	 << endl;
	    tCapacity-=(last->getEndNodeDemand()+first->getStartNodeDemand());
	    //cout << " tCapacity " << tCapacity 
	    //   << " last->getStartNodeDemand() " << last->getStartNodeDemand()
            //	 << endl;
	    tCapacity+=(last->getStartNodeDemand());
	    //cout << " tCapacity " << tCapacity << endl; 

	    replaceElementAt(0,eEdge);
            for(i=eList->getSize()-2;i>=0;i--)   
	      if((last=eList->getNthEdge(i))!=NULL) {
	         addElementFront(*last);
	         dCost+=last->getCost(); 
	         tCapacity+=(last->getStartNodeDemand());
		 //cout << " dCost " << dCost
		 //     << " last->getCost() " << last->getCost()
		 //     << " last->getStartNodeDemand() " << last->getStartNodeDemand()
		 //     << " tCapacity " << tCapacity; 
	      }// end of if

	     ret = 1;

	      if(!isValid(eGraph->getProblem(),1)) {
		//cout << "res: " << *this << endl;
	        cout << " res " << endl;
                printTabular(cout);
                eList->printTabular(cout);
		//cout <<" dummy0 " << dummy0 << " dummy1 " << dummy1 << endl;
		cout <<" 1. EdgeList.mergeWith() " << endl;
		exit(1);
              }// end of if
            */

	    EdgeListPtr rt = EdgeList::merge2EdgeList(this,eList,eGraph->getProblem());
	    if(rt!=NULL) { ret=1; *this=*rt; delete rt; }

            if(!isValid(eGraph->getProblem(),1)) {
		//cout << "res: " << *this << endl;
	        cout << " res " << endl;
                printTabular(cout);
                eList->printTabular(cout);
		//cout <<" dummy0 " << dummy0 << " dummy1 " << dummy1 << endl;
		cout <<" 1. EdgeList.mergeWith() " << endl;
		exit(1);
              }// end of if

	}// end of if
    } else if( eGraph!=NULL && eList->getStartDate()>getEndDate() )
           {
              EdgePtr last  = getLastEdge();
              EdgePtr first = eList->getFirstEdge();
              // If there is an edge to combine them
              if( last!=NULL && first!=NULL && 
                  eGraph->doesEdgeExist(last->getStartNode(),
					first->getEndNode(),
					last->getStartDate())) 
	      {  //cout << " mergeWith " << endl;
		 //printTabular(cout); cout << " and " << endl;
		 //eList->printTabular(cout); 
		 // Becareful, two edges are replaced by one edge.
	         vrpcost dummy0 = first->getCost();
	         vrpcost dummy1 = last->getCost();
	         Edge eEdge(eGraph,last->getStartNode(),first->getEndNode(),last->getStartDate());
	         dCost += (eEdge.getCost()-dummy1);
	         tCapacity-=(last->getEndNodeDemand()+first->getStartNodeDemand());
	         tCapacity+=(eEdge.getEndNodeDemand());
		 wtime+=(eEdge.getWaitingTime()-last->getWaitingTime());

	         replaceElementAt(getSize()-1,eEdge);
		 vrptime currTime = eEdge.getStartDateSecondNode();
                 for(i=1;i<eList->getSize();i++)   
	           if((last=eList->getNthEdge(i))!=NULL) {
		     //Edge a(eGraph,last->getStartNode(),last->getEndNode(),currTime);
	             /*addElement(*last);
	             dCost+=last->getCost(); 
	             tCapacity+=(last->getEndNodeDemand());
		     */
		     eEdge=*last;
		     eEdge.setStartDate(currTime);
		     addEdge(eEdge);

	             //addElement(eEdge);
	             //dCost+=eEdge.getCost(); 
	             //tCapacity+=(eEdge.getEndNodeDemand());
		     //wtime+=eEdge.getWaitingTime();
		     currTime = eEdge.getStartDateSecondNode();
	           }// end of if

	         ret = 1;

	      }// end of if

	      if(!isValid(eGraph->getProblem(),1)) {
		cout <<" 2. EdgeList.mergeWith() " << endl;
		exit(1);
              }// end of if

           }// end of if

     else if( eGraph!=NULL ) {
	    EdgeListPtr rt = EdgeList::merge2EdgeList(this,eList,eGraph->getProblem());
	    if(rt!=NULL) { ret=1; *this=*rt; delete rt; }
	  }

   //if(ret) {
   //  ofstream out("canmerge.out",ios::app);
   //  out <<" merged " << *this << endl;
   //  out.close();
   //}// end of if

   return ret;

}// end of mergeWith()

/*************************************************************************/
/**  complete merge operation 
  1. both of them should be NOT NULL
  2. Total demand should be LESS THAN or EQUAL to the maximum Vehicle
     Capacity.
*/
/*************************************************************************/
EdgeListPtr 
EdgeList::merge2EdgeList(EdgeListPtr eList1,EdgeListPtr eList2,VRProblem* prob)
{
   EdgeListPtr ret=NULL;
   int i,m;
   vrptime startD;  
   vrptime currTime;  
   vrptime endD;  
   vrptime t1,t2;  
   EdgePtr edge1;
   EdgePtr edge2;
   EdgePtr edge;
   EdgePtr ePtr;
   customerId currNode, cNode1, cNode2;
   Customer cus1;
   Customer cus2;
   char aaaa;
   EdgeGraph *eGraph = prob->getEdgeGraph(); 

   if( prob!=NULL && eList1!=NULL && eList2!=NULL &&
       eList1->getNumberOfEdges()>0 && eList2->getNumberOfEdges()>0 && 
       (eList1->getDemand()+eList2->getDemand())<=prob->getVehicleCapacity() )
   {
      ret = new EdgeList();
      ePtr = eList1->getNthEdge(0);
      i=1; m=1; currNode = ePtr->getStartNode();
      currTime=ePtr->getStartDate();
      int valid = 1;

      while(valid&&i<eList1->getNumberOfEdges()&&m<eList2->getNumberOfEdges()) {
	edge1 = eList1->getNthEdge(i);
	edge2 = eList2->getNthEdge(m);
	cNode1= edge1->getStartNode();
	cNode2= edge2->getStartNode();
	prob->getCustomer(cNode1,cus1);
	prob->getCustomer(cNode2,cus2);
        t1=cus1.getStartTime();
        t2=cus2.getStartTime();
	if(t1<t2) {
	  if( eGraph->doesEdgeExist(currNode,cNode1,currTime ) ) {
	    edge = new Edge(eGraph,currNode,cNode1,currTime);
	    currNode=cNode1; i++; 
	  } else valid=0;
	} else {
	    if( eGraph->doesEdgeExist(currNode,cNode2,currTime ) ) {
	      edge = new Edge(eGraph,currNode,cNode2,currTime);
	      currNode=cNode2; m++; 
	    } else valid=0;
	  }
	if(valid) {
	  currTime=edge->getStartDateSecondNode();
	  ret->addEdge(*edge);
	  delete edge;
	}// end of if
      }// end of while

      while(valid&&i<eList1->getNumberOfEdges()) {
	edge1 = eList1->getNthEdge(i);
	cNode1= edge1->getStartNode();
	if( eGraph->doesEdgeExist(currNode,cNode1,currTime ) ) {
	  edge = new Edge(eGraph,currNode,cNode1,currTime);
	  currNode=cNode1; i++; 
	  currTime=edge->getStartDateSecondNode();
	  ret->addEdge(*edge);
	  delete edge;
	} else valid=0;
      }// end of while

      while(valid&&m<eList2->getNumberOfEdges()) {
	edge2 = eList2->getNthEdge(m);
	cNode2= edge2->getStartNode();
	if( eGraph->doesEdgeExist(currNode,cNode2,currTime ) ) {
	  edge = new Edge(eGraph,currNode,cNode2,currTime);
	  currNode=cNode2; m++; 
	  currTime=edge->getStartDateSecondNode();
	  ret->addEdge(*edge);
	  delete edge;
	} else valid=0;
      }// end of while

      if(valid&&eGraph->doesEdgeExist(currNode,edge1->getEndNode(),currTime)) {
        cNode1= edge1->getEndNode();
        edge = new Edge(eGraph,currNode,cNode1,currTime);
        ret->addEdge(*edge);
        delete edge;
      } else valid=0;

      if(valid) {
        if(!ret->isValid(prob,1))  { delete ret; ret=NULL; }
        /*else {
	      cout <<" merge2EdgeList " << endl
		   << "eList1 " << *eList1 << endl
		   << "eList2 " << *eList2 << endl
		   << "res : "  << *ret << endl;


	   }*/ 
      } else { delete ret; ret=NULL; }

   }// end of if

   return ret;

}// end of merge2EdgeList()

/*************************************************************************/
/** 
    Returns the number of different elements.
**/
/*************************************************************************/
EdgeList::getNumberOfDifferences(EdgeListPtr eList)
{
   int i, j, ret = 0;
   int *nodes  = new int[getSize()+1];
   int *nodes1 = new int[eList->getSize()+1];
   Edge *edge=NULL;

   for(i=0;i<getSize();i++)
      if((edge=getNthEdge(i))!=NULL) nodes[i]=edge->getStartNode();
   if( edge!=NULL ) nodes[i]=edge->getEndNode();

   edge=NULL;
   for(i=0;i<eList->getSize();i++)
      if((edge=eList->getNthEdge(i))!=NULL) nodes1[i]=edge->getStartNode();
   if( edge!=NULL ) nodes1[i]=edge->getEndNode();

   i=j=0;
   while( i<=getSize() && j<=eList->getSize() ) 
      if( nodes[i]<nodes1[j] ) { ret++; i++; }
      else if( nodes[i]>nodes1[j] ) j++; 
      else { i++; j++; }

   ret += (getSize()-i+1);

   delete[] nodes;
   delete[] nodes1;

   return ret;

}// end of getNumberOfDifferences()

/*************************************************************************/
/*************************************************************************/
EdgeListPtr EdgeList::combineEdgeLists(EdgeListPtr first,  unsigned fIndex, 
				       EdgeListPtr second, unsigned sIndex, 
				       VRProblemPtr prob)
{  int k;
   EdgeListPtr temp = new EdgeList();
   EdgeGraphPtr edgeGraph = prob->getEdgeGraph();
   EdgePtr edge1 = first->getNthEdge(fIndex);
   EdgePtr edge2 = second->getNthEdge(sIndex);
   int fr1 = edge1->getStartNode();
   int to2 = edge2->getEndNode();
   Customer cus;

   k=0;
   while(k<fIndex) { temp->addEdge(*first->getNthEdge(k)); k++; }

   Edge prev(edgeGraph,fr1,to2,edge1->getStartDate());
   temp->addEdge(prev);

   k=sIndex+1;
   while(k<second->getNumberOfEdges()) 
   {  
     EdgePtr kthEdge = second->getNthEdge(k);
     prob->getCustomer(kthEdge->getStartNode(),cus);
     Edge anEdge(edgeGraph,kthEdge->getStartNode(),kthEdge->getEndNode(),
  	         maximum(prev.getStartDateSecondNode(),cus.getStartTime()));
     temp->addEdge(anEdge);

     prev=anEdge;
     k++;

   }// end of while

   if(!temp->isValid(prob,1)) {
      cout << " combineEdgeLists not VALID! " << endl;
      temp->printTabular(cout);
      exit(1);
   }// end of if

   return temp;

}// end of combineEdgeLists()

/*************************************************************************/
/*************************************************************************/
EdgeListPtr 
EdgeList::combineEdgeLists(EdgeListPtr first, int f0, int f1,
			 EdgeListPtr second,int s0,int s1,VRProblemPtr prob ) {
   int k;
   EdgeListPtr temp = new EdgeList();
   EdgeGraphPtr edgeGraph = prob->getEdgeGraph();
   EdgePtr edge1 = first->getNthEdge(f0);
   EdgePtr edge2 = second->getNthEdge(s0);
   int fr1 = edge1->getStartNode();
   int to2 = edge2->getEndNode();
   Customer cus;
   Edge anEdge;

   k=0;
   while(k<f0) { temp->addEdge(*first->getNthEdge(k)); k++; }

   Edge prev(edgeGraph,fr1,to2,edge1->getStartDate());
   temp->addEdge(prev);

   k=s0+1;
   while(k<s1&&k<second->getNumberOfEdges()) {
     EdgePtr kthEdge = second->getNthEdge(k);
     prob->getCustomer(kthEdge->getStartNode(),cus);
     Edge aEdge(edgeGraph,kthEdge->getStartNode(),kthEdge->getEndNode(),
                maximum(prev.getStartDateSecondNode(),cus.getStartTime()));
     anEdge=aEdge;
     temp->addEdge(anEdge);
     prev=anEdge; k++;
   }// end of while

   edge1 = first->getNthEdge(f1);
   edge2 = second->getNthEdge(s1);
   fr1 = edge2->getStartNode();
   to2 = edge1->getEndNode();

   Edge aPrev(edgeGraph,fr1,to2,prev.getStartDateSecondNode());
   temp->addEdge(aPrev);
   prev=aPrev;

   k=f1+1;
   while(k<first->getNumberOfEdges()) {
     EdgePtr kthEdge = first->getNthEdge(k);
     prob->getCustomer(kthEdge->getStartNode(),cus);
     Edge aEdge(edgeGraph,kthEdge->getStartNode(),kthEdge->getEndNode(),
                maximum(prev.getStartDateSecondNode(),cus.getStartTime()));
     anEdge=aEdge;
     temp->addEdge(anEdge);
     prev=anEdge; k++;
   }// end of while

   if(!temp->isValid(prob,1)) {
      cout << " combineEdgeLists_V2 not VALID! " << endl;
      temp->printTabular(cout);
      exit(1);
   }// end of if

   return temp;

}// end of combineEdgeLists()

/*************************************************************************/
/*************************************************************************/
EdgeList::splitEdgeList( vrptime onePX, EdgeListPtr* r0, EdgeListPtr* r1,
			       VRProblemPtr prob)
{ int k;
  int retValue = 0;
  EdgePtr edge;
  //cout << " EdgeList::splitEdgeList " << *this << endl;
  if( onePX<getStartDate() ) {
    *r0 = new EdgeList(*this);
    retValue = 1;
  } 
  else if( onePX>getEndDate() ) {
    *r1 = new EdgeList(*this);
    retValue=2;
  } else {
    *r0 = new EdgeList();
    *r1 = new EdgeList();
    EdgeGraph* eGraph = prob->getEdgeGraph();
    int i=0, stay=1;
    int currNode = getStartNode();
    vrptime startD = getStartDate();
    retValue = 3;

    while( i<getNumberOfEdges() && stay) {
      edge = getNthEdge(i);
      if( edge!=NULL && edge->getEndDate()<onePX ) {
        Edge anEdge(eGraph,currNode,edge->getEndNode(),startD);
        (*r0)->addEdge(anEdge);
        startD = anEdge.getStartDateSecondNode();
        currNode = anEdge.getEndNode();
        i++;  
      } else stay=0; // end of if
    }// end of while 

    if(currNode!=getStartNode()) {
      Edge anEdge1(eGraph,currNode,getStartNode(),startD);
      (*r0)->addEdge(anEdge1);
    }// end of if
    edge = getNthEdge(i);
    currNode = edge->getEndNode();

    if(currNode!=getStartNode()) {
      Edge anEdge2(eGraph,getStartNode(),currNode,getStartDate());
      (*r1)->addEdge(anEdge2);
      i++; 
      startD=anEdge2.getStartDateSecondNode();
      while(i<getNumberOfEdges()) {
        if( (edge = getNthEdge(i))!=NULL ) {
          Edge anEdge3(eGraph,currNode,edge->getEndNode(),startD);
          (*r1)->addEdge(anEdge3);
          currNode = anEdge3.getEndNode();
          startD = anEdge3.getStartDateSecondNode();
        }// end of if
        i++;
      }// end of while 
    }// end of if

    if( (*r1)->getNumberOfEdges()<2 ) { delete *r1; *r1=NULL; retValue-=2; }
    if( (*r0)->getNumberOfEdges()<2 ) { delete *r0; *r0=NULL; retValue-=1; }
  }

  return retValue;

}// end of splitEdgeList()

/*************************************************************************/
/*************************************************************************/
EdgeList::splitEdgeList( vrptime onePX, vrptime twoPX, 
                         EdgeListPtr* r0, EdgeListPtr* r1, EdgeListPtr* r2,
			       VRProblemPtr prob)
{ int k;
  int retValue = 0;
  EdgePtr edge;
  cout << " EdgeList::splitEdgeList " << endl;
  if( onePX>getEndDate() ) {
    *r0 = new EdgeList(*this);
    retValue = 1;
  } else if( twoPX<getStartDate() ) {
    *r2 = new EdgeList(*this);
    retValue=4;
  } else if( onePX>getStartDate() && twoPX>getEndDate() ) {
    retValue = splitEdgeList(onePX,r0,r1,prob);
  } else if( onePX<getStartDate() && twoPX<getEndDate() ) {
    retValue = splitEdgeList(twoPX,r1,r2,prob)*2;
  } else {
    *r0 = new EdgeList();
    *r1 = new EdgeList();
    *r2 = new EdgeList();
    retValue = 7;
    EdgeGraph* eGraph = prob->getEdgeGraph();
    int i=0, stay=1;
    int currNode = getStartNode();
    vrptime startD = getStartDate();

    while( i<getNumberOfEdges() && stay) {
      edge = getNthEdge(i);
      if( edge!=NULL && edge->getEndDate()<onePX ) {
        Edge anEdge(eGraph,currNode,edge->getEndNode(),startD);
        (*r0)->addEdge(anEdge);
        startD = anEdge.getStartDateSecondNode();
        currNode = anEdge.getEndNode();
        i++;  
      } else stay=0; // end of if
    }// end of while 
    
    if(currNode!=getStartNode()) { 
      Edge anEdge1(eGraph,currNode,getStartNode(),startD);
      (*r0)->addEdge(anEdge1);
    }// end of if
    
    edge = getNthEdge(i);
    currNode = edge->getEndNode();
    Edge anEdge2(eGraph,getStartNode(),currNode,getStartDate());
    (*r1)->addEdge(anEdge2);
    i++; startD=anEdge2.getStartDateSecondNode();

    while(i<getNumberOfEdges()) {
      if( (edge = getNthEdge(i))!=NULL && edge->getEndDate()<twoPX ) {
        Edge anEdge3(eGraph,currNode,edge->getEndNode(),startD);
        (*r1)->addEdge(anEdge3);
        currNode = anEdge3.getEndNode();
        startD = anEdge3.getStartDateSecondNode();
      }// end of if
      i++;
    }// end of while
    Edge anEdge11(eGraph,currNode,getStartNode(),startD);
    (*r1)->addEdge(anEdge11);

    edge = getNthEdge(i);
    currNode = edge->getEndNode();
    if(currNode!=getStartNode()) {
      Edge anEdge21(eGraph,getStartNode(),currNode,getStartDate());
      (*r2)->addEdge(anEdge21);
      i++; startD=anEdge21.getStartDateSecondNode();
      while(i<getNumberOfEdges()) {
        if( (edge = getNthEdge(i))!=NULL ) {
          Edge anEdge3(eGraph,currNode,edge->getEndNode(),startD);
          (*r2)->addEdge(anEdge3);
          currNode = anEdge3.getEndNode();
          startD = anEdge3.getStartDateSecondNode();
        }// end of if
        i++;
      }// end of while
    }// end of if

    if( (*r2)->getNumberOfEdges()<2 ) { delete *r2; *r2=NULL; retValue-=4; }
    if( (*r1)->getNumberOfEdges()<2 ) { delete *r1; *r1=NULL; retValue-=2; }
    if( (*r0)->getNumberOfEdges()<2 ) { delete *r0; *r0=NULL; retValue-=1; }
  }

  return retValue;

}// end of splitEdgeList()

/*************************** T H E   E N D *******************************/
